# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from .general import _pkg_root, __version__, citation

from .beam_elements import BlackAbsorber, BlackCrystal, TransparentCollimator, TransparentCrystal, \
                           EverestBlock, EverestCollimator, EverestCrystal, BlowUp, EmittanceMonitor, \
                           collimator_classes, crystal_classes, element_classes
from .scattering_routines.everest import materials, Material, CrystalMaterial
from .colldb import CollimatorDatabase
from .interaction_record import InteractionRecord
from .rf_sweep import RFSweep, prepare_rf_sweep
from .lossmap import LossMap, MultiLossMap


# print("If you use Xcoll in your simulations, please cite us :-)")
# print(citation)




# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #


LOST_ON_EVEREST_BLOCK   = -330
LOST_ON_EVEREST_COLL    = -331
LOST_ON_EVEREST_CRYSTAL = -332

LOST_ON_FLUKA_BLOCK     = -333
LOST_ON_FLUKA_COLL      = -334
LOST_ON_FLUKA_CRYSTAL   = -335

LOST_ON_GEANT4_BLOCK    = -336
LOST_ON_GEANT4_COLL     = -337
LOST_ON_GEANT4_CRYSTAL  = -338

LOST_ON_ABSORBER        = -340

MASSLESS_OR_NEUTRAL     = -341
ACC_IONISATION_LOSS     = -342  # Not a real particle, but accumulation of some ionisation losses that are not accounted for (per collimator)
VIRTUAL_ENERGY          = -343  # Not a real particle, but energy that is deposited

ERR_INVALID_TRACK       = -390
ERR_NOT_IMPLEMENTED     = -391
ERR_INVALID_XOFIELD     = -392
ERR                     = -399

particle_states_src = f"""
#ifndef XCOLL_STATES_H
#define XCOLL_STATES_H
#define  XC_LOST_ON_EVEREST_BLOCK   {LOST_ON_EVEREST_BLOCK}
#define  XC_LOST_ON_EVEREST_COLL    {LOST_ON_EVEREST_COLL}
#define  XC_LOST_ON_EVEREST_CRYSTAL {LOST_ON_EVEREST_CRYSTAL}
#define  XC_LOST_ON_FLUKA_BLOCK     {LOST_ON_FLUKA_BLOCK}
#define  XC_LOST_ON_FLUKA_COLL      {LOST_ON_FLUKA_COLL}
#define  XC_LOST_ON_FLUKA_CRYSTAL   {LOST_ON_FLUKA_CRYSTAL}
#define  XC_LOST_ON_GEANT4_BLOCK    {LOST_ON_GEANT4_BLOCK}
#define  XC_LOST_ON_GEANT4_COLL     {LOST_ON_GEANT4_COLL}
#define  XC_LOST_ON_GEANT4_CRYSTAL  {LOST_ON_GEANT4_CRYSTAL}
#define  XC_LOST_ON_ABSORBER        {LOST_ON_ABSORBER}
#define  XC_MASSLESS_OR_NEUTRAL     {MASSLESS_OR_NEUTRAL}
#define  XC_ACC_IONISATION_LOSS     {ACC_IONISATION_LOSS}
#define  XC_VIRTUAL_ENERGY          {VIRTUAL_ENERGY}
#define  XC_ERR_INVALID_TRACK       {ERR_INVALID_TRACK}
#define  XC_ERR_NOT_IMPLEMENTED     {ERR_NOT_IMPLEMENTED}
#define  XC_ERR_INVALID_XOFIELD     {ERR_INVALID_XOFIELD}
#define  XC_ERR                     {ERR}
#endif /* XCOLL_STATES_H */
"""


# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import os
import numpy as np
from numbers import Number
from functools import wraps

import xobjects as xo
import xtrack as xt
import xtrack.particles.pdg as pdg

try:
    # TODO: once xaux is in Xsuite keep only this
    from xaux import FsPath, ranID
except (ImportError, ModuleNotFoundError):
    from ..xaux import FsPath, ranID


class BaseEngine(xo.HybridClass):
    _xofields = {
        '_particle_ref': xt.Particles._XoStruct,
        '_seed':         xo.UInt64,
        '_capacity':     xo.Int64,
    }

    _int32 = False
    _only_protons = False
    _element_classes = None
    _uses_input_file = False
    _num_input_files = 0
    _uses_run_folder = False

    def __init__(self, **kwargs):
        if self._element_classes is None:
            raise NotImplementedError(f"{self.__class__.__name__} needs to define `_element_classes`!")
        # Initialise defaults
        self._cwd = None
        self._line = None
        self._verbose = False
        self._input_file = None
        self._element_dict = {}
        self._warning_given = False
        self._environment = None
        self._element_index = 0
        self._tracking_initialised = False
        self._deactivated_elements = {}
        kwargs.setdefault('_particle_ref', xt.Particles())
        kwargs.setdefault('_seed', 0)
        kwargs.setdefault('_capacity', 0)
        super().__init__(**kwargs)

    def __del__(self, *args, **kwargs):
        self.stop(warn=False)

    def _warn(self, error=None):
        if not self._warning_given:
            print(f"Warning: Failed to import {self.__class__.__name__} environment "
                + f"(did you compile?).\n{self.name.capitalize()} elements can be installed "
                + f"but are not trackable.", flush=True)
            if error:
                print(f"Error: {error}", flush=True)
            self._warning_given = True
        self.stop()
        if error:
            raise error

    def _print(self, *args, **kwargs):
        if self.verbose:
            kwargs.setdefault('flush', True)
            print(*args, **kwargs)


    # ==================
    # === Properties ===
    # ==================

    @property
    def environment(self):
        return self._environment

    @property
    def name(self):
        return self.__class__.__name__.replace('Engine', '').lower()

    @property
    def verbose(self):
        return self._verbose

    @verbose.setter
    def verbose(self, val):
        self._verbose = val

    @property
    def line(self):
        return self._line

    @line.setter
    def line(self, val):
        if not val is None and not isinstance(val, xt.Line):
            raise ValueError("`line` has to be an xt.Line object!")
        self._line = val

    @line.deleter
    def line(self):
        self._line = None

    @property
    def particle_ref(self):
        initial = xt.Particles().to_dict()
        current = self._particle_ref.to_dict()
        if xt.line._dicts_equal(initial, current):
            return None
        else:
            return self._particle_ref

    @particle_ref.setter
    def particle_ref(self, val):
        if val is None:
            self._particle_ref = xt.Particles()
        else:
            if not isinstance(val, xt.Particles):
                raise ValueError("`particle_ref` has to be an xt.Particles object!")
            if val._capacity > 1:
                raise ValueError("`particle_ref` has to be a single particle!")
            pdg_id = val.pdg_id[0]
            if pdg_id == 0:
                if self._only_protons:
                    pdg_id = pdg.get_pdg_id_from_name('proton')
                else:
                    raise ValueError(f"{self.__class__.__name__} allows the use of particles "
                                   + f"different than protons. Hence, `particle_ref` "
                                   + f"needs to have a valid pdg_id.")
            elif self._only_protons and pdg_id != pdg.get_pdg_id_from_name('proton'):
                raise ValueError("{self.__class__.__name__} only supports protons!")
            self._particle_ref = val
            self._particle_ref.pdg_id[0] = pdg_id

    @particle_ref.deleter
    def particle_ref(self):
        self.particle_ref = None

    @property
    def capacity(self):
        if self._capacity == 0:
            return None
        else:
            return int(self._capacity)

    @capacity.setter
    def capacity(self, val):
        if val is None:
            val = 0
        if not isinstance(val, Number) or val < 0:
            raise ValueError("`capacity` has to be a positive integer!")
        self._capacity = int(val)

    @capacity.deleter
    def capacity(self):
        self.capacity = None

    @property
    def seed(self):
        if self._seed == 0:
            return None
        else:
            return self._seed

    @seed.setter
    def seed(self, val):
        if val is None:
            val = 0
        if not isinstance(val, Number) or val < 0:
            raise ValueError("`seed` has to be a positive integer!")
        val = int(val)
        if self._int32:
            new_val = np.uint32(val)
        else:
            new_val = np.uint64(val)
        if new_val != val:
            self._print(f"Warning: type change for seed {val}. Using {new_val}.")
        self._seed = new_val

    @seed.deleter
    def seed(self):
        self.seed = None

    @property
    def input_file(self):
        if self._uses_input_file:
            return self._input_file

    @property
    def element_dict(self):
        return self._element_dict


    # ======================
    # === Public Methods ===
    # ======================

    def start(self, *, clean=True, input_file=None, **kwargs):
        if not self.environment:
            raise RuntimeError(f"{self.name.capitalize()} environment not set up! "
                             + f"Do not manually create an instance of the engine.")
        if not self.environment.initialised:
            raise RuntimeError(f"{self.name.capitalize()} environment not initialised! "
                             + f"Please set all paths in the environment before "
                             + f"starting the engine.")
        if not self.environment.compiled:
            raise RuntimeError(f"{self.name.capitalize()} interface not compiled! "
                             + f"Please compile before starting the engine.")
        if self.is_running():
            self._print("Engine already running.")
            return

        # Clean up any leftover failed runs
        self.stop(clean=clean)

        print(f"Starting {self.__class__.__name__}...   ", flush=True, end='')
        kwargs = self._pre_start(**kwargs)

        # This needs to be set in the ChildEngine, either in _start_engine() or at the start of tracking
        self._tracking_initialised = False

        # Set all engine properties that have a setter (this will remove these properties from the kwargs)
        kwargs = self._set_engine_properties(**kwargs)

        # Create input file if needed (this will remove the kwargs relevant to the input file and physics)
        kwargs = self._use_input_file(input_file, **kwargs)
        if clean:
            self.clean_input_files(clean_all=False)
        self._preparing_input = False

        # Start the engine in the ChildEngine
        self._start_engine(**kwargs)

        # Done starting
        if self.verbose:
            print(f"{self.__class__.__name__} started.", flush=True)
        else:
            print(f"Done.", flush=True)

    def stop(self, clean=False, **kwargs):
        kwargs = self._stop_engine(**kwargs)
        if clean:
            self.clean(clean_all=True, **kwargs)
        self._restore_engine_properties(clean=clean)
        self._warning_given = False
        self._tracking_initialised = False


    def assert_particle_ref(self):
        if self.particle_ref is None:
            raise ValueError(f"{self.__class__.__name__} reference particle not set!")


    def generate_input_file(self, *, clean=True, filename=None, **kwargs):
        # This method manually generates an input file without starting the engine
        if not self._uses_input_file:
            raise ValueError(f"{self.__class__.__name__} does not use input files!")
        if self._element_dict:
            raise ValueError("Elements already assigned to engine (cannot regenerate input "
                           + "file after starting engine)!")

        # Set all engine properties that have a setter (this will remove these properties from the kwargs)
        kwargs = self._set_engine_properties(**kwargs)

        # Create input file
        input_file, _ = self._generate_input_file(**kwargs)
        if not hasattr(input_file, '__iter__') or isinstance(input_file, str):
            # Some engines might need multiple input files (like Fluka)
            input_file = [input_file]
        if filename is None:
            if hasattr(self, '_old_cwd') and self._old_cwd is not None:
                input_file[0] = input_file[0].rename(self._old_cwd / input_file[0].name)
        else:
            input_file[0] = input_file[0].rename(filename)
        for i, file in enumerate(input_file[1:]):
            input_file[i+1] = file.rename(input_file[0].parent / file.name)
        input_file = input_file[0] if self._num_input_files==1 else input_file

        if clean:
            self.clean_input_files(clean_all=False)
        self._restore_engine_properties(clean=clean)

        return input_file


    def is_running(self):
        if hasattr(self, '_preparing_input') and self._preparing_input:
            # We need this to allow changing the element settings which otherwise are locked
            return False
        # If we get here, we cannot say if the engine is running or not and we need an
        # implementation in the child class
        return self._is_running()


    def clean(self, **kwargs):
        self.clean_input_files(**kwargs)
        self.clean_output_files(**kwargs)

    def clean_input_files(self, **kwargs):
        kwargs = self._get_input_cwd_for_cleaning(**kwargs)
        self._clean_input_files(**kwargs)

    def clean_output_files(self, **kwargs):
        kwargs = self._get_input_cwd_for_cleaning(**kwargs)
        self._clean_output_files(**kwargs)


    # =======================
    # === Private Methods ===
    # =======================

    # For all the engine fields, they can either be set in advance on the engine,
    # or they can be set when the engine is started. In the latter case, the values
    # are temporary and the original will be restored when the engine is stopped.

    def _set_engine_properties(self, **kwargs):
        self._preparing_input = True  # We need this to allow changing the element settings which otherwise are locked
        def _set_property(prop):
            val = kwargs.pop(prop, None)
            if val is not None:
                # We only need to update the property when it is not None
                setattr(self, f'_old_{prop}', getattr(self, prop))
                setattr(self, prop, val)
        # We need to set the following properties first as they are needed by the others
        _set_property('verbose')
        _set_property('line')
        # The following properties have a specific logic
        self._use_seed(kwargs.pop('seed', None))
        self._use_particle_ref(kwargs.pop('particle_ref', None))
        self._sync_line_particle_ref()
        self._get_elements(kwargs.pop('elements', None), kwargs.pop('names', None))
        self._set_cwd(kwargs.pop('cwd', None))
        # Now we can set the rest of the properties
        _set_property('capacity')
        return kwargs

    def _restore_engine_properties(self, clean=False):
        self._preparing_input = False
        # Reset particle_ref in the line
        if hasattr(self, '_old_particle_ref'):
            self.line.particle_ref = self._old_line_particle_ref
            del self._old_line_particle_ref
        # The following properties have a specific logic
        self._reactivate_elements()
        self._reset_cwd(clean=clean)
        # Reset all other properties
        self_attributes = self.__dict__.copy()
        for kk, vv in self_attributes.items():
            if kk.startswith('_old_'):
                prop = kk[5:]
                setattr(self, prop, vv)
                delattr(self, kk)
        self._input_file = None
        self._element_dict = {}

    def _use_seed(self, seed=None):
        if seed is None:
            if self.seed is None:
                if self._int32:
                    self.seed = np.random.randint(0, int(2**32))
                else:
                    self.seed = np.random.randint(0, int(2**64))
        else:
            self._old_seed = self.seed
            self.seed = seed
        self._print(f"Using seed {self.seed}.")

    def _use_particle_ref(self, particle_ref=None):
        # Prefer: provided particle_ref > existing particle_ref > particle_ref from line
        if particle_ref is not None:
            self._old_particle_ref = self.particle_ref
            self.particle_ref = particle_ref
        elif self.particle_ref is None:
            if self.line is None or not hasattr(self.line, 'particle_ref') \
            or self.line.particle_ref is None:
                raise ValueError("Need to provide either a line with a reference "
                               + "particle, or `particle_ref`.")
            self._old_particle_ref = self.particle_ref
            self.particle_ref = self.line.particle_ref
        self._print(f"Using {pdg.get_name_from_pdg_id(self.particle_ref.pdg_id[0])} "
                  + f"with momentum {self.particle_ref.p0c[0]/1.e9:.1f} GeV.")

    def _sync_line_particle_ref(self):
        if self.line is None:
            return
        if self.line.particle_ref is not None \
        and not xt.line._dicts_equal(self.line.particle_ref.to_dict(),
                                     self.particle_ref.to_dict()):
            self._print("Found different reference particle in line. Temporarily overwritten.")
            self._old_line_particle_ref = self.line.particle_ref
            self.line.particle_ref = self.particle_ref

    def _get_new_element_name(self):
        name = f"{self.name}_el_{self._element_index}"
        self._element_index += 1
        return name

    def _deactivate_element(self, el):
        self._deactivated_elements[el.name] = [el, el.active or True]
        if hasattr(el, 'active'):
            el.active = False
        self._remove_element(el)

    def _assert_element(self, element):
        if not isinstance(element, self._element_classes):
            raise ValueError(f"Element {element.name} is not a "
                            + ", or a ".join([c.__name__ for c in self._element_classes])
                            + ".")

    def _get_elements(self, elements=None, names=None):
        if elements is not None and (not hasattr(elements, '__iter__') or isinstance(elements, str)):
            elements = [elements]
        if names is not None and (not hasattr(names, '__iter__') or isinstance(names, str)):
            names = [names]
        if self.line is None:
            if elements is None:
                raise ValueError("Need to provide either `line` or `elements`.")
            if names is None:
                names = []
                for ee in elements:
                    if hasattr(ee, 'name'):
                        names.append(ee.name)
                    else:
                        name = self._get_new_element_name()
                        names.append(name)
                        ee.name = name
            elif len(names) == len(elements):
                for ee, name in zip(elements, names):
                    if hasattr(ee, 'name'):
                        if ee.name != name:
                            self._print(f"Warning: Element name {ee.name} changed to {name}.")
                            ee.name = name
            else:
                raise ValueError("Length of `elements` and `names` doesn't match.")
        else:
            if elements is not None:
                raise ValueError("Cannot provide both `line` and `elements`.")
            if names is None:
                elements, names = self.line.get_elements_of_type(self._element_classes)
            else:
                elements = [self.line[name] for name in names]
        this_names = []
        this_elements = []
        for ee, name in zip(elements, names):
            self._assert_element(ee)
            if ee.jaw is None:
                self._print(f"Warning: Jaw not set for {name}. Ignoring.")
                self._deactivate_element(ee)
            elif not ee.active:
                self._print(f"Warning: Element {name} is not active. Ignoring.")
                self._deactivate_element(ee)
            else:
                this_names.append(name)
                this_elements.append(ee)
        if len(this_elements) == 0:
            raise ValueError(f"No active {self.name} elements found!")
        self._element_dict = dict(zip(this_names, this_elements))

    def _reactivate_elements(self):
        names = list(self._deactivated_elements.keys())
        for name in names:
            ee, was_active = self._deactivated_elements.pop(name)
            self._restore_element(ee)
            if hasattr(ee, 'active'):
                ee.active = was_active

    def _set_cwd(self, cwd=None):
        if self._uses_run_folder:
            if cwd is not None:
                cwd = FsPath(cwd).expanduser().resolve()
                if cwd.exists() and cwd != FsPath.cwd():
                    # Check if the folder already exists
                    # If the specified folder is the current one, we do not need to rename
                    i = 0
                    while (cwd.parent / f'{cwd.name}_{i:0>4}').exists():
                        i += 1
                        if i > 9999:
                            raise ValueError(f"Too many folders with the same "
                                           + f"name {cwd}!")
                    cwd = cwd.parent / f'{cwd.name}_{i:0>4}'
            else:
                ran_str = ranID(only_alphanumeric=True)
                cwd = FsPath.cwd() / f'{self.name}_run_{ran_str}'
            self._cwd = cwd
            cwd.mkdir(parents=True, exist_ok=True)
            self._old_cwd = FsPath.cwd()
            os.chdir(cwd)

    def _reset_cwd(self, clean=False):
        if hasattr(self, '_old_cwd'):
            if self._old_cwd is not None:
                os.chdir(self._old_cwd)
                if clean and self._cwd is not None and self._cwd != FsPath.cwd():
                    try:
                        self._cwd.rmdir()
                    except:
                        self._print(f"Warning: Failed to remove temporary folder "
                                  + f"{self._cwd}.")
            del self._old_cwd
        self._cwd = None


    def _use_input_file(self, input_file=None, **kwargs):
        if self._uses_input_file:
            if input_file is None:
                input_file, kwargs = self._generate_input_file(**kwargs)
            if not hasattr(input_file, '__iter__') or isinstance(input_file, str):
                # Some engines might need multiple input files (like Fluka)
                input_file = [input_file]
            input_file = [FsPath(f).expanduser().resolve() for f in input_file]
            new_files = []
            for file in input_file:
                if not file.exists():
                    raise ValueError(f"Input file {file.as_posix()} not found!")
                if file.parent != FsPath.cwd() and self._uses_run_folder:
                    file.copy_to(FsPath.cwd(), method='mount')
                    new_files.append(FsPath.cwd() / file.name)
                else:
                    new_files.append(file)
            self._input_file = new_files[0] if self._num_input_files==1 else new_files
            self._match_input_file()
        return kwargs


    def _get_input_cwd_for_cleaning(self, **kwargs):
        # Get the input file and the cwd
        input_file = kwargs.get('input_file', None)
        cwd = kwargs.get('cwd', None)
        if input_file is None:
            if self.input_file is not None:
                input_file = self.input_file
            if cwd is None and self._cwd is not None:
                cwd = self._cwd
        else:
            if not hasattr(input_file, '__iter__') or isinstance(input_file, str):
                input_file = [input_file]
            input_file = [FsPath(ff) for ff in input_file]
            if cwd is None:
                cwd = input_file[0].parent
            if len(input_file) == 1:
                input_file = input_file[0]
        kwargs['input_file'] = input_file
        if self._uses_run_folder:
            kwargs['cwd'] = cwd
        else:
            kwargs['cwd'] = FsPath.cwd()
        return kwargs


    # =================================================
    # === Methods to be overwritten by child engine ===
    # =================================================

    def _start_engine(self, **kwargs):
        raise NotImplementedError(f"Need to implement `_start_engine` for "
                                   + f"{self.__class__.__name__}!")

    def _stop_engine(self, **kwargs):
        raise NotImplementedError(f"Need to implement `_stop_engine` for "
                                   + f"{self.__class__.__name__}!")

    def _is_running(self, **kwargs):
        raise NotImplementedError(f"Need to implement `_is_running` for "
                                   + f"{self.__class__.__name__}!")

    def _match_input_file(self, **kwargs):
        if self._uses_input_file:
            raise NotImplementedError(f"Need to implement `_match_input_file` for "
                                   + f"{self.__class__.__name__}!")

    def _generate_input_file(self, **kwargs):
        if self._uses_input_file:
            raise NotImplementedError("Need to implement `_generate_input_file` for "
                                   + f"{self.__class__.__name__}!")


    # ============================================================
    # === Methods to be optionally overwritten by child engine ===
    # ============================================================

    def _clean_input_files(self, input_file=None, cwd=None, clean_all=False, **kwargs):
        pass

    def _clean_output_files(self, input_file=None, cwd=None, clean_all=False, **kwargs):
        pass

    def _remove_element(self, el):
        pass

    def _restore_element(self, el):
        pass

    def _pre_start(self, **kwargs):
        return kwargs



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import numpy as np
from numbers import Number


_NORMS = ["total", "coll_max", "max", "none", "raw"]


def _plot_lossmap_base(lossmap: dict, *, norm="total", ax=None, xlim=None, ylim=None,
                       legend=True, grid=True, energy=False, show=True, savefig=None):
    import matplotlib.pyplot as plt
    if norm in _NORMS:
        norm = norm.lower()
    elif not isinstance(norm, Number):
        raise ValueError(f"Norm must be one of {_NORMS} or a number, not {norm}.")

    coll_s = lossmap['collimator']['s']
    coll_val = lossmap['collimator']['e'] if energy else lossmap['collimator']['n']
    if lossmap['interpolation']:
        cold_s = lossmap['aperture']['s_bins']
        cold_val = lossmap['aperture']['e_bins'] if energy else lossmap['aperture']['n_bins']
    else:
        cold_s = lossmap['aperture']['s']
        cold_val = lossmap['aperture']['e'] if energy else lossmap['aperture']['n']
    warm_s = np.array([])  # Placeholder for warm losses, if needed
    warm_val = np.array([])  # Placeholder for warm losses, if needed
    if norm != "raw":
        if lossmap['interpolation']:
            cold_val = cold_val / lossmap['interpolation']
            # warm_val = warm_val / lossmap['interpolation']
        coll_val = coll_val / lossmap['collimator']['length']
        if norm == "total":
            scale = coll_val.sum() + cold_val.sum() + warm_val.sum()
        elif norm == "coll_max":
            scale = coll_val.max()
        elif norm == "max":
            scale = max(coll_val.max(), cold_val.max(), warm_val.max())
        elif norm == "none":
            scale = 1
        else:
            scale = norm
        coll_val = coll_val / scale
        cold_val = cold_val / scale
        warm_val = warm_val / scale
        if lossmap['interpolation']:
            cold_val = cold_val / lossmap['aperture']['length_bins']

    L = lossmap['machine_length']
    xlim = xlim if xlim else [-0.01*L, 1.01*L]
    ylim = ylim if ylim else [1.e-7, 1.e1]

    font = {'family': 'serif', 'size': 17}
    format_dict = {f"font.{prop}": font[prop] for prop in font}
    with plt.rc_context(format_dict):
        if ax is None:
            _, ax = plt.subplots(figsize=(16, 4))

        if grid:
            ax.grid(axis="y", which="major", zorder=0)

        bar_common_kwargs = dict(width = 0.8, lw = 1, bottom = 1.e-9)
        ax.bar(coll_s, coll_val, color="k", edgecolor="k", label="Collimator", zorder=10, **bar_common_kwargs)
        ax.bar(cold_s, cold_val, color="b", edgecolor="b", label="Cold",  zorder=11, **bar_common_kwargs)
        # ax.bar(warm_s, warm_val, color="r", edgecolor="r", label="Warm",  zorder=12, **bar_common_kwargs)

        ax.set_yscale("log")
        ax.set_ylim(ylim)
        ax.set_xlim(xlim)

        ax.set_yticks([10**i for i in range(int(np.log10(ylim[0])), 1 + int(np.log10(ylim[1])))])
        ax.set_xlabel("s [m]")
        ax.set_ylabel(_label_from_norm(norm, energy))

        if legend:
            ax.legend(loc='best', fancybox=True, framealpha=0.8, fontsize="small")

        plt.tight_layout()
        if savefig:
            plt.savefig(savefig, dpi=300, bbox_inches='tight')
        if show:
            plt.show()
        else:
            plt.close()
        return ax.figure, ax


def _label_from_norm(norm, energy):
    if norm == "none":
        if energy:
            return "Deposited energy [eV/m]"
        else:
            return "Particles absorbed [1/m]"
    elif norm == "raw":
        if energy:
            return "Deposited energy [eV]"
        else:
            return "Particles absorbed [-]"
    elif norm == "total":
        return "Norm. inefficiency [1/m]"
    elif norm == "max":
        return "Inefficiency (norm. by max) [-]"
    elif norm == "coll_max":
        return "Inefficiency (norm. by coll max) [-]"
    else:
        return f"Inefficiency (norm. by {norm}) [1/m]"



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

from .base import BaseBlock, BaseCollimator, BaseCrystal
from .absorber import BlackAbsorber, BlackCrystal
from .transparent import TransparentCollimator, TransparentCrystal
from .everest import EverestBlock, EverestCollimator, EverestCrystal
from .blowup import BlowUp
from .monitor import EmittanceMonitor

block_classes = tuple(v for v in globals().values()
                      if isinstance(v, type) and issubclass(v, BaseBlock) and v != BaseBlock
                      and v != BaseCollimator and v != BaseCrystal)
# Includes crystals
collimator_classes = tuple(v for v in globals().values()
                           if isinstance(v, type) and (issubclass(v, BaseCollimator) or issubclass(v, BaseCrystal))
                           and v != BaseCollimator and v != BaseCrystal)
crystal_classes = tuple(v for v in globals().values()
                        if isinstance(v, type) and issubclass(v, BaseCrystal) and v != BaseCrystal)

element_classes = block_classes + (BlowUp, EmittanceMonitor)



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import numpy as np
import pandas as pd
from pathlib import Path
import json

import xtrack as xt
import xobjects as xo

from .beam_elements import collimator_classes, crystal_classes
from .general import __version__
from .plot import _plot_lossmap_base
from .headers.particle_states import (
    LOST_ON_EVEREST_BLOCK, LOST_ON_EVEREST_COLL, LOST_ON_EVEREST_CRYSTAL,
    LOST_ON_FLUKA_BLOCK, LOST_ON_FLUKA_COLL, LOST_ON_FLUKA_CRYSTAL,
    LOST_ON_GEANT4_BLOCK, LOST_ON_GEANT4_COLL, LOST_ON_GEANT4_CRYSTAL,
    LOST_ON_ABSORBER, MASSLESS_OR_NEUTRAL, ACC_IONISATION_LOSS,
    VIRTUAL_ENERGY)


class LossMap:
    def __init__(self, line, part, *, line_is_reversed, interpolation=None,
                 line_shift_s=0, weights=None, weight_function=None, verbose=True):
        self._line_is_reversed = line_is_reversed
        self._machine_length = line.get_length() if line else None
        self._interpolation = interpolation
        self._momentum = line.particle_ref.p0c[0] if line else None
        self._aper_s = np.array([])
        self._aper_nabs = np.array([])
        self._aper_eabs = np.array([])
        self._aperbins = np.array([])
        self._aperbins_length = np.array([])
        self._aperbinned = np.array([])
        self._aperbinned_energy = np.array([])
        self._coll_s = np.array([])
        self._coll_name = np.array([])
        self._coll_nabs = np.array([])
        self._coll_eabs = np.array([])
        self._coll_length = np.array([])
        self._coll_type = np.array([])
        self._xcoll = __version__
        self._energy_waring_given = False
        if part and line:
            self.add_particles(part=part, line=line, line_shift_s=line_shift_s,
                               weights=weights, weight_function=weight_function,
                               verbose=verbose)

    def __str__(self):
        return f"LossMap ({self.num_aperture_losses} losses on aperture and " \
             + f"{self.num_collimator_losses} losses on collimators)"

    def __repr__(self):
        return f"<{str(self)} at {hex(id(self))}>"

    @classmethod
    def from_json(cls, file, verbose=True):
        lm = cls(None, None, line_is_reversed=None)
        lm.add_from_json(file, verbose=verbose)
        return lm

    def to_json(self, file):
        with open(Path(file), 'w') as fid:
            json.dump({
                'xcoll': self._xcoll,
                'date':  pd.Timestamp.now().strftime('%Y-%m-%d %H:%M:%S'),
                **self.lossmap
            }, fid, indent=True, cls=xo.JEncoder)

    def save_summary(self, file):
        with open(Path(file), 'w') as fid:
            fid.write(self.summary.__repr__())


    @property
    def lossmap(self):
        coll_summary = self.summary[self.summary.n > 0].to_dict('list')
        coll_summary = {kk: np.array(vv) for kk, vv in coll_summary.items()}
        return {
                'collimator':      coll_summary,
                'aperture':        self.aperture_losses,
                'machine_length':  self.machine_length,
                'interpolation':   self.interpolation,
                'reversed':        self.line_is_reversed,
                'momentum':        self._momentum,
            }

    @property
    def summary(self):
        return pd.DataFrame({
            'name':   self._coll_name,
            'n':      self._coll_nabs,
            'e':      self._coll_eabs,
            'length': self._coll_length,
            's':      self._coll_s,
            'type':   self._coll_type,
        }).sort_values("s")

    @property
    def aperture_losses(self):
        if self._interpolation:
            mask = self._aperbinned > 0
            aper_s = self._aperbins[:-1] + self.interpolation/2
            return {
                'idx_bins': np.arange(len(self._aperbins)-1)[mask],
                's_bins': aper_s[mask],
                'n_bins': self._aperbinned[mask],
                'e_bins': self._aperbinned_energy[mask],
                'length_bins': self._aperbins_length[mask],
            }
        else:
            return {
                's': self._aper_s,
                'n': self._aper_nabs,
                'e': self._aper_eabs
            }

    @property
    def line_is_reversed(self):
        return self._line_is_reversed

    @property
    def machine_length(self):
        return self._machine_length

    @property
    def interpolation(self):
        return self._interpolation

    @property
    def momentum(self):
        return self._momentum

    @property
    def num_losses(self):
        return self.num_aperture_losses + self.num_collimator_losses

    @property
    def num_aperture_losses(self):
        return int(self._aperbinned.sum() + self._aper_nabs.sum())

    @property
    def num_collimator_losses(self):
        return int(self._coll_nabs.sum())


    def plot(self, *, norm="total", ax=None, xlim=None, ylim=None, legend=True,
             grid=True, energy=False, show=True, savefig=None):
        return _plot_lossmap_base(self.lossmap, norm=norm, ax=ax, xlim=xlim,
                            ylim=ylim, legend=legend, grid=grid, energy=energy,
                            show=show, savefig=savefig)

    def add_particles(self, part, line, *, line_shift_s=0, weights=None,
                      weight_function=None, verbose=True):
        """
        Add particles to the loss map. Aperture losses are interpolated and the
        collimator summary is updated.
        """
        if self.interpolation is None:
            self._interpolation = 0.1
        if self._machine_length:
            if not np.isclose(self._machine_length, line.get_length()):
                raise ValueError("The line length is different from the one used "
                                 "to create the loss map.")
        else:
            self._machine_length = line.get_length()
        if self._momentum:
            if not np.isclose(self._momentum, line.particle_ref.p0c[0]):
                raise ValueError("The reference momentum is different from the one "
                                 "used to create the loss map.")
        else:
            self._momentum = line.particle_ref.p0c[0]
        if weights is None:
            if weight_function is None:
                weights = np.ones(len(part.x))
            else:
                weights = _create_weights_from_initial_state(part, weight_function)
        else:
            if weight_function is not None:
                raise ValueError("Use either 'weights' or 'weight_function', not both!")
            if len(weights) != len(part.x):
                raise ValueError("The length of the weights array must be equal to "
                                "the number of particles.")

        if line_shift_s != 0:
            raise NotImplementedError("Line shift not implemented yet.")

        # Correct particles that are lost in aperture directly after collimator.
        # These should be absorbed.
        self._correct_absorbed(part, line, verbose=verbose)

        # Loss location refinement
        if self._interpolation:
            self._interpolate(part, line, verbose=verbose)

        self._make_coll_summary(part, line, line_shift_s, weights)
        self._get_aperture_losses(part, line_shift_s, weights)


    def add_from_json(self, file, verbose=True):
        """
        Add loss map data from a JSON file or an iterable of files. The loss map
        is updated with the data from the file(s).
        """
        if hasattr(file, '__iter__') and not isinstance(file, (str, bytes, Path)):
            i = 0
            for ff in file:
                self.add_from_json(ff)
                i += 1
            if verbose:
                print(f"Loaded {i} files into loss map.")
            return
        with open(Path(file), 'r') as fid:
            lossmap = json.load(fid)
        new_style_json = LossMap._assert_valid_json(lossmap)
        if new_style_json:
            if self._momentum is None:
                self._momentum = lossmap['momentum']
            elif not np.isclose(self._momentum, lossmap['momentum']):
                raise ValueError("The momentum is different from the one used "
                                 "to create the loss map.")
            if self._xcoll != lossmap['xcoll'] and verbose:
                print("Warning: The xcoll version is different from the one used "
                    "to create the loss map.")
        if self.line_is_reversed is None:
            self._line_is_reversed = lossmap['reversed']
        elif self.line_is_reversed != lossmap['reversed']:
            raise ValueError("The line_is_reversed is different from the one used "
                             "to create the loss map.")
        if self.machine_length is None:
            self._machine_length = lossmap['machine_length']
        elif not np.isclose(self.machine_length, lossmap['machine_length']):
            raise ValueError("The line length is different from the one used "
                             "to create the loss map.")
        if self.interpolation is None:
            self._interpolation = lossmap['interpolation']
        elif not np.isclose(self.interpolation, lossmap['interpolation']):
            raise ValueError("The interpolation step is different from the one used "
                             "to create the loss map.")
        self._load_coll_summary(lossmap['collimator'])
        self._load_aperture_losses(lossmap['aperture'])


    def _correct_absorbed(self, part, line, verbose=True):
        # Correct particles that are at an aperture directly after a collimator
        coll_classes = list(set(collimator_classes) - set(crystal_classes))
        coll_elements = line.get_elements_of_type(coll_classes)[1]
        for idx, elem in enumerate(part.at_element):
            if part.state[idx] == 0:
                if elem == 0:
                    prev_elem = len(line.element_names) - 1
                else:
                    prev_elem = elem - 1
                if line.element_names[prev_elem] in coll_elements:
                    if verbose:
                        print(f"Found at {line.element_names[elem]}, "
                            + f"moved to {line.element_names[elem-1]}")
                    part.at_element[idx] = elem - 1
                    what_type = line[elem-1].__class__.__name__
                    if what_type == 'EverestBlock':
                        part.state[idx] = LOST_ON_EVEREST_BLOCK
                    elif what_type == 'EverestCollimator':
                        part.state[idx] = LOST_ON_EVEREST_COLL
                    elif what_type == 'EverestCrystal':
                        part.state[idx] = LOST_ON_EVEREST_CRYSTAL
                    elif what_type == 'FlukaBlock':
                        part.state[idx] = LOST_ON_FLUKA_BLOCK
                    elif what_type == 'FlukaCollimator':
                        part.state[idx] = LOST_ON_FLUKA_COLL
                    elif what_type == 'FlukaCrystal':
                        part.state[idx] = LOST_ON_FLUKA_CRYSTAL
                    elif what_type == 'Geant4Block':
                        part.state[idx] = LOST_ON_GEANT4_BLOCK
                    elif what_type == 'Geant4Collimator':
                        part.state[idx] = LOST_ON_GEANT4_COLL
                    elif what_type == 'Geant4Crystal':
                        part.state[idx] = LOST_ON_GEANT4_CRYSTAL
                    elif what_type == 'BlackAbsorber':
                        part.state[idx] = LOST_ON_ABSORBER
                    else:
                        raise ValueError(f"Unknown collimator type {what_type}")


    def _interpolate(self, part, line, verbose=True):
        if len(part.s[part.state==0]) > 0:
            if verbose:
                print("Performing the aperture losses refinement.")
            loss_loc_refinement = xt.LossLocationRefinement(
                line,
                n_theta = 360,            # Angular resolution
                r_max = 0.5,              # Maximum transverse aperture [m]
                dr = 50e-6,               # Transverse accuracy [m]
                ds = self.interpolation   # Longitudinal accuracy [m]
            )
            loss_loc_refinement.refine_loss_location(part)


    def _make_coll_summary(self, part, line, line_shift_s, weights):
        all_lost_states = [LOST_ON_EVEREST_BLOCK, LOST_ON_EVEREST_COLL, LOST_ON_EVEREST_CRYSTAL,
                           LOST_ON_FLUKA_BLOCK, LOST_ON_FLUKA_COLL, LOST_ON_FLUKA_CRYSTAL,
                           LOST_ON_GEANT4_BLOCK, LOST_ON_GEANT4_COLL, LOST_ON_GEANT4_CRYSTAL,
                           LOST_ON_ABSORBER, MASSLESS_OR_NEUTRAL, ACC_IONISATION_LOSS,
                           VIRTUAL_ENERGY
        ]

        names = line.get_elements_of_type(collimator_classes)[1]
        coll_mask = (part.state <= max(all_lost_states)) & (part.state >= min(all_lost_states))
        coll_losses = np.array([line.element_names[i]
                                  for i in part.at_element[coll_mask]])
        coll_lengths = [line[name].length for name in names]

        L = self.machine_length
        coll_pos = [(line.get_s_position(name) + cl/2 + line_shift_s)%L
                    for name, cl in zip(names, coll_lengths)]
        if self.line_is_reversed:
            coll_pos = L - coll_pos

        coll_types = [line[name].__class__.__name__
                      for name in names]
        coll_weights = weights[coll_mask]
        nabs = [coll_weights[coll_losses == name].sum()
                for name in names]
        energy_weights = coll_weights * part.energy[coll_mask]
        eabs = [energy_weights[coll_losses == name].sum()
                for name in names]
        self._do_collimator_adding(coll_s=coll_pos, coll_name=names, coll_nabs=nabs,
                                   coll_eabs=eabs, coll_length=coll_lengths,
                                   coll_type=coll_types)

    def _load_coll_summary(self, colldata):
        coll_eabs = colldata['e'] if 'e' in colldata else np.zeros(len(colldata['s']))
        if 'type' in colldata:
            coll_types = colldata['type']
        else:
            coll_types = np.full(len(colldata['s']), "Unknown", dtype=object)
        self._do_collimator_adding(coll_s=colldata['s'], coll_name=colldata['name'],
                                   coll_nabs=colldata['n'], coll_eabs=coll_eabs,
                                   coll_length=colldata['length'], coll_type=coll_types)

    def _do_collimator_adding(self, coll_s, coll_name, coll_nabs, coll_eabs,
                              coll_length, coll_type):
        # TODO: this can be done smarter, masks instead of loops, though
        # one should keep in mind that indices can be repeated and then
        # masking does not work correctly
        for ss, nn, nabs, eabs, ll, tt in zip(coll_s, coll_name, coll_nabs,
                                        coll_eabs, coll_length, coll_type):
            idx, = np.where(self._coll_name == nn)
            if len(idx) == 0:
                self._coll_s = np.append(self._coll_s, ss)
                self._coll_name = np.append(self._coll_name, nn)
                self._coll_nabs = np.append(self._coll_nabs, nabs)
                self._coll_eabs = np.append(self._coll_eabs, eabs)
                self._coll_length = np.append(self._coll_length, ll)
                self._coll_type = np.append(self._coll_type, tt)
            else:
                idx = idx[0]
                if not np.isclose(self._coll_length[idx], ll):
                    raise ValueError(f"Length of {nn} is different from the one used "
                                    + "to create the loss map.")
                if self._coll_type[idx] != tt and tt != 'Unknown':
                    raise ValueError(f"Type of {nn} is different from the one used to "
                                    + "create the loss map.")
                if not np.isclose(self._coll_s[idx], ss):
                    raise ValueError(f"s position of {nn} is different from the one "
                                    + "used to create the loss map.")
                self._coll_nabs[idx] += nabs
                self._coll_eabs[idx] += eabs


    def _make_aperture_bins(self):
        if self._aperbins.size == 0:
            L = self.machine_length
            end = L - np.mod(L, self.interpolation) + self.interpolation
            self._aperbins = np.linspace(0, end, int(np.ceil(L/self.interpolation)) + 1)
            # TODO: not entirely correct, as not all bins have the same lenght in practice: If
            # an active element like a magnet starts in the middle of the bin, that bin is
            # essentially shorter. Then the first aperture marker after the element represents
            # the particles that would be lost in this element, so there the aperture length is
            # equal to the element length. This is not true if the active element is a collimator
            # (then the losses are moved to the collimator anyway).
            # TODO: need length for non-interpolated case as well?
            self._aperbins_length = np.full(len(self._aperbins) - 1, self.interpolation, dtype=np.float64)
            self._aperbinned = np.zeros(len(self._aperbins) - 1, dtype=np.float64)
            self._aperbinned_energy = np.zeros(len(self._aperbins) - 1, dtype=np.float64)

    def _get_aperture_losses(self, part, line_shift_s, weights):
        aper_mask = part.state == 0
        if len(part.s[aper_mask]) == 0:
            return

        # Get s position per particle (lost on aperture)
        L = self.machine_length
        aper_s = np.mod(part.s[aper_mask] + line_shift_s, L)
        if self._line_is_reversed:
            aper_s = L - aper_s

        if self._interpolation:
            # Binned aperture losses
            self._make_aperture_bins()
            self._do_aperture_binning(aper_s=aper_s, aper_nabs=weights[aper_mask],
                                      aper_eabs=weights[aper_mask] * part.energy[aper_mask])
        else:
            # Aperture losses at exact s positions (because no interpolation performed)
            # TODO: need correct lengths to scale
            aper_pos       = np.unique(aper_s)
            aper_weights   = weights[aper_mask]
            aper_nabs      = [aper_weights[aper_s == j].sum() for j in aper_pos]
            energy_weights = aper_weights * part.energy[aper_mask]
            aper_eabs      = [energy_weights[aper_s == s].sum() for s in aper_pos]
            self._do_aperture_adding(aper_s=aper_pos, aper_nabs=aper_nabs,
                                     aper_eabs=aper_eabs)

    def _load_aperture_losses(self, aperdata):
        if self._interpolation:
            self._make_aperture_bins()
            if not 'idx_bins' in aperdata and 's' in aperdata and 'n' in aperdata:
                # Old-style JSON file
                aper_eabs = aperdata['e'] if 'e' in aperdata else np.zeros(len(aperdata['s']))
                self._do_aperture_binning(aper_s=aperdata['s'], aper_nabs=aperdata['n'],
                                          aper_eabs=aper_eabs)
            else:
                self._aperbinned[aperdata['idx_bins']] = aperdata['n_bins']
                self._aperbinned_energy[aperdata['idx_bins']] = aperdata['e_bins']
        else:
            self._do_aperture_adding(aper_s=aperdata['s'], aper_nabs=aperdata['n'],
                                     aper_eabs=aperdata['e'])

    def _do_aperture_binning(self, aper_s, aper_nabs, aper_eabs):
        binned = np.digitize(aper_s, bins=self._aperbins, right=False) - 1
        # We cannot directly use binned as an index, as repeated
        # indices would be ignored
        minlength = len(self._aperbins) - 1
        self._aperbinned += np.bincount(binned, weights=aper_nabs,
                                        minlength=minlength)
        self._aperbinned_energy += np.bincount(binned, weights=aper_eabs,
                                               minlength=minlength)

    def _do_aperture_adding(self, aper_s, aper_nabs, aper_eabs):
        # TODO: this can be done smarter, masks instead of loops, though
        # one should keep in mind that indices can be repeated and then
        # masking does not work correctly
        for ss, nabs, eabs in zip(aper_s, aper_nabs, aper_eabs):
            idx, = np.where(self._aper_s == ss)
            if len(idx) == 0:
                self._aper_s = np.append(self._aper_s, ss)
                self._aper_nabs = np.append(self._aper_nabs, nabs)
                self._aper_eabs = np.append(self._aper_eabs, eabs)
            else:
                idx = idx[0]
                self._aper_nabs[idx] += nabs
                self._aper_eabs[idx] += eabs


    @staticmethod
    def _assert_valid_json(lossmap):
        is_new_format = 'xcoll' in lossmap
        if 'machine_length' not in lossmap:
            raise ValueError("The JSON file does not contain the machine length data.")
        if 'interpolation' not in lossmap:
            raise ValueError("The JSON file does not contain the interpolation data.")
        if 'reversed' not in lossmap:
            raise ValueError("The JSON file does not contain the reversed data.")
        if is_new_format:
            if 'momentum' not in lossmap:
                raise ValueError("The JSON file does not contain the momentum data.")
        if 'collimator' not in lossmap:
            raise ValueError("The JSON file does not contain the collimator data.")
        if 's' not in lossmap['collimator']:
            raise ValueError("The JSON file does not contain the collimator s data.")
        if 'name' not in lossmap['collimator']:
            raise ValueError("The JSON file does not contain the collimator name data.")
        if 'length' not in lossmap['collimator']:
            raise ValueError("The JSON file does not contain the collimator length data.")
        if 'n' not in lossmap['collimator']:
            raise ValueError("The JSON file does not contain the collimator n data.")
        if is_new_format:
            if 'e' not in lossmap['collimator']:
                raise ValueError("The JSON file does not contain the collimator energy data.")
            if 'type' not in lossmap['collimator']:
                raise ValueError("The JSON file does not contain the collimator type data.")
        if 'aperture' not in lossmap:
            raise ValueError("The JSON file does not contain the aperture data.")
        if is_new_format and lossmap['interpolation'] is not None:
            if 'idx_bins' not in lossmap['aperture']:
                raise ValueError("The JSON file does not contain the aperture idx_bins data.")
            if 'n_bins' not in lossmap['aperture']:
                raise ValueError("The JSON file does not contain the aperture n_bins data.")
            if 'e_bins' not in lossmap['aperture']:
                raise ValueError("The JSON file does not contain the aperture e_bins data.")
            if 'length_bins' not in lossmap['aperture']:
                raise ValueError("The JSON file does not contain the aperture length_bins data.")
        else:
            if 's' not in lossmap['aperture']:
                raise ValueError("The JSON file does not contain the aperture s data.")
            if 'n' not in lossmap['aperture']:
                raise ValueError("The JSON file does not contain the aperture n data.")
            if is_new_format:
                if 'e' not in lossmap['aperture']:
                    raise ValueError("The JSON file does not contain the aperture e data.")
                if 'length' not in lossmap['aperture']:
                    raise ValueError("The JSON file does not contain the aperture length data.")
        return is_new_format


class MultiLossMap(LossMap):
    def __init__(self, *lms, **named_lms):
        """
        Create a MultiLossMap object from a list of LossMap objects.
        """
        super().__init__(None, None, line_is_reversed=None)
        del self._line_is_reversed
        self._lms = []
        self._lm_names = []
        if len(lms) > 0:
            if len(named_lms) > 0:
                raise ValueError("Specify the LossMaps as args or kwargs, not both.")
            if hasattr(lms[0], '__iter__'):
                if len(lms) > 1:
                    raise ValueError("Use args or a list of LossMap objects, not both.")
                if isinstance(lms[0], dict):
                    lms = lms[0].values()
                    self._lm_names = lms[0].keys()
                else:
                    lms = lms[0]
        elif len(named_lms) > 0:
            lms = named_lms.values()
            self._lm_names = named_lms.keys()
        for lm in lms:
            self.add_lossmap(lm)

    @property
    def lms(self):
        if len(self._lm_names) > 0:
            return dict(zip(self._lm_names, self._lms))
        else:
            return self._lms

    @property
    def line_is_reversed(self):
        raise AttributeError("The line_is_reversed property is not available "
                           + "for MultiLossMap objects.")

    @property
    def lossmap(self):
        coll_summary = self.summary[self.summary.n > 0].to_dict('list')
        coll_summary = {kk: np.array(vv) for kk, vv in coll_summary.items()}
        return {
                'collimator':      coll_summary,
                'aperture':        self.aperture_losses,
                'machine_length':  self.machine_length,
                'interpolation':   self.interpolation,
                'momentum':        self._momentum,
            }


    def add_lossmap(self, lm):
        """
        Add a LossMap object to the MultiLossMap object.
        """
        if not isinstance(lm, LossMap):
            raise ValueError("The input must be a LossMap object.")
        self._lms.append(lm)
        if self._machine_length is None:
            self._machine_length = lm.machine_length
        elif not np.isclose(self._machine_length, lm.machine_length):
            raise ValueError("The machine lengths of the loss maps are not the same.")
        if self._momentum is None:
            self._momentum = lm.momentum
        elif lm.momentum and not np.isclose(self._momentum, lm.momentum):
            raise ValueError("The momenta of the loss maps are not the same.")
        if self._interpolation is None:
            self._interpolation = lm.interpolation
        elif not np.isclose(self._interpolation, lm.interpolation):
            raise ValueError("The interpolations of the loss maps are not the same.")
        self._do_collimator_adding(coll_s=lm._coll_s, coll_name=lm._coll_name,
                                   coll_nabs=lm._coll_nabs, coll_eabs=lm._coll_eabs,
                                   coll_length=lm._coll_length, coll_type=lm._coll_type)
        if self.interpolation:
            if len(lm._aperbins) == 0:
                # No aperture in this loss map
                return
            elif len(self._aperbins) == 0:
                self._aperbins = lm._aperbins
                self._aperbins_length = lm._aperbins_length
                self._aperbinned = lm._aperbinned
                self._aperbinned_energy = lm._aperbinned_energy
            else:
                if len(self._aperbins) != len(lm._aperbins) \
                or not np.allclose(self._aperbins, lm._aperbins):
                    raise ValueError("The number of bins of the loss maps are not the same.")
                self._aperbinned += lm._aperbinned
                self._aperbinned_energy += lm._aperbinned_energy
        else:
            self._do_aperture_adding(aper_s=lm._aper_s, aper_nabs=lm._aper_nabs,
                                     aper_eabs=lm._aper_eabs)


def _create_weights_from_initial_state(part, function):
    if len(function) == 4:
        return function[0](part.x)*function[1](part.px)*\
               function[2](part.y)*function[3](part.py)
    elif len(function) == 6:
        return function[0](part.x)*function[1](part.px)*\
               function[2](part.y)*function[3](part.py)*\
               function[4](part.zeta)*function[5](part.delta)
    else:
        raise NotImplementedError




# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import os
import sys
import json
import tempfile
from subprocess import run, PIPE
# from platformdirs import user_config_dir, user_data_dir

from ..general import _pkg_root
try:
    from xaux import FsPath  # TODO: once xaux is in Xsuite keep only this
except (ImportError, ModuleNotFoundError):
    from ..xaux import FsPath


class BaseEnvironment:
    _pkg_root
    _config_dir = (_pkg_root / 'config').resolve()
    _data_dir   = (_pkg_root / 'lib').resolve()
    # _config_dir = FsPath(user_config_dir('xcoll')).resolve()
    # _data_dir   = FsPath(user_data_dir('xcoll')).resolve()
    _paths = {} # The value is the parent depth that needs to be brute-forced (0 = file itself, None = no brute-force)
    _read_only_paths = {}

    def __init__(self, *args, **kwargs):
        self._old_sys_path = None
        self._old_os_env = None
        self._temp_dir = None
        for path in self._paths.keys():
            setattr(self, f'_{path}', None)
        self._config_file = self._config_dir / f'{self.__class__.__name__[:-11].lower()}.config.json'
        sys.path.append(self._data_dir.as_posix())
        self._in_constructor = True
        self.load()
        self._in_constructor = False

    def __del__(self):
        self.restore_environment()
        if self._temp_dir:
            self._temp_dir.cleanup()

    def __repr__(self):
        return f"<{self.__class__.__name__} at {hex(id(self))} (use .show() to see the paths)>"

    def __str__(self):
        res = ["XcollEnvironment"]
        res.append(f"    Configuration file:  {self._config_file.as_posix()}")
        res.append(f"    Configuration dir:   {self._config_dir.as_posix()}")
        res.append(f"    Data dir:            {self._data_dir.as_posix()}")
        if self._temp_dir:
            res.append(f"    Temporary dir:       {self._temp_dir.name}")
        if self.__class__ is not BaseEnvironment:
            res.append("")
            res.append(f"{self.__class__.__name__}")
            for path in self._paths.keys():
                value = getattr(self, f'_{path}', None)
                path = f'{path}:'
                if value is None:
                    res.append(f"    {path:<20} None")
                else:
                    res.append(f"    {path:<20} {value.as_posix()}")
            for path in self._read_only_paths.keys():
                value = getattr(self, path)
                path = f'{path} (read-only):'
                res.append(f"    {path:<20} {value.as_posix()}")
            if self._old_sys_path and self._old_os_env:
                res.append("")
                res.append("Custom environment stored:")
                res.append(f"    sys.path: {self._old_sys_path}")
                res.append(f"    os.environ: {self._old_os_env}")
        return "\n".join(res)

    @property
    def config_file(self):
        """The environment configuration file."""
        return self._config_file

    @property
    def config_dir(self):
        """The directory where the configuration files are stored."""
        return self._config_dir

    @property
    def data_dir(self):
        """The directory where the data files are stored."""
        return self._data_dir

    @property
    def initialised(self):
        return all(getattr(self, path, None) is not None and getattr(self, path, None).exists()
                   for path in self._paths.keys())

    @property
    def compiled(self):
        raise NotImplementedError("This property should be implemented in the subclass.")

    @property
    def temp_dir(self):
        if not self._temp_dir:
            self._temp_dir = tempfile.TemporaryDirectory()
        return FsPath(self._temp_dir.name)

    @temp_dir.setter
    def temp_dir(self, value):
        if value is None:
            if self._temp_dir:
                self._temp_dir.cleanup()
                self._temp_dir = None
            return
        if self._temp_dir:
            self._temp_dir.cleanup()
            self._temp_dir = None
        if isinstance(value, FsPath):
            value = value.resolve()
        elif isinstance(value, str):
            value = FsPath(value).resolve()
        else:
            raise TypeError("temp_dir must be a string or FsPath.")
        if not value.exists():
            raise FileNotFoundError(f"Provided temp_dir {value} does not exist!")
        self._temp_dir = tempfile.TemporaryDirectory(dir=value)

    @temp_dir.deleter
    def temp_dir(self):
        self.temp_dir = None

    def show(self):
        """Print the environment paths."""
        print(self)

    def save(self):
        if not self._config_dir.exists():
            self._config_dir.mkdir(parents=True, exist_ok=True)
        data = {}
        for path in self._paths.keys():
            value = getattr(self, path, None)
            if value:
                value = FsPath(value).as_posix()
            data[path] = value
        with open(self._config_file, 'w') as fid:
            json.dump(data, fid, indent=4)

    def load(self):
        if not self._config_file.exists():
            self.save()
        with open(self._config_file, 'r') as fid:
            data = json.load(fid)
        for key, value in data.items():
            setattr(self, key, FsPath(value) if value else None)

    def store_environment(self):
        self._old_sys_path = sys.path.copy()
        self._old_os_env = os.environ.copy()

    def restore_environment(self):
        if self._old_sys_path:
            sys.path = self._old_sys_path
            self._old_sys_path = None
        if self._old_os_env:
            os.environ = self._old_os_env
            self._old_os_env = None

    def brute_force_path(self, path):
        num_parents = 0
        if path in self._paths:
            num_parents = self._paths[path]
            path = getattr(self, path)
        if path in self._read_only_paths:
            num_parents = self._read_only_paths[path]
            path = getattr(self, path)
        if path is None:
            return
        path = FsPath(path).resolve()
        if num_parents > 0:
            path = path.parents[num_parents-1]
        if not path.exists():
            raise FileNotFoundError(f"Could not find path {path}!")
        try:
            cmd = run(['tree', path.as_posix()], stdout=PIPE, stderr=PIPE)
        except FileNotFoundError:
            # No tree executable. Return as no more can be done. TODO: alternatives for mac and windows?
            return
        if cmd.returncode != 0:
            stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
            raise RuntimeError(f"Could not resolve {path} tree!\nError given is:\n{stderr}")

    def __getattr__(self, key):
        if key in self._paths.keys():
            value = getattr(self, f'_{key}', None)
            if value:
                return FsPath(value)
        else:
            raise AttributeError(f"{self.__class__.__name__} has no attribute '{key}'")

    def __setattr__(self, key, value):
        if key in self._paths.keys():
            if value:
                value = FsPath(value)
                if not self._in_constructor:
                    self.brute_force_path(value)
            setattr(self, f'_{key}', value)
            if not self._in_constructor:
                self.save()
        elif key in self._read_only_paths.keys():
            raise AttributeError(f"Attribute '{key}' of {self.__class__.__name__} "
                               + f"is read-only!")
        else:
            super().__setattr__(key, value)

    def __delattr__(self, item):
        if item in self._paths.keys():
            self.__setattr__(self, f'_{item}', None)
            self.save()

    def assert_make_installed(self, verbose=False):
        if hasattr(self, '_make_installed') and hasattr(self, '_cmake_installed'):
            return self._make_installed and self._cmake_installed
        try:
            cmd = run(["make", "--version"], stdout=PIPE, stderr=PIPE)
        except FileNotFoundError:
            self._make_installed = False
            raise RuntimeError("Could not find make installation!")
        if cmd.returncode == 0:
            self._make_installed = True
        else:
            stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
            self._make_installed = False
            raise RuntimeError(f"Could not run make! Verify its installation.\nError given is:\n{stderr}")
        try:
            cmd = run(["cmake", "--version"], stdout=PIPE, stderr=PIPE)
        except FileNotFoundError:
            self._cmake_installed = False
            raise RuntimeError("Could not find cmake installation!")
        if cmd.returncode == 0:
            self._cmake_installed = True
        else:
            stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
            self._cmake_installed = False
            raise RuntimeError(f"Could not run cmake! Verify its installation.\nError given is:\n{stderr}")

    def assert_gcc_isntalled(self, verbose=False):
        if hasattr(self, '_gcc_installed'):
            return self._gcc_installed
        try:
            cmd = run(["gcc", "-dumpversion"], stdout=PIPE, stderr=PIPE)
        except FileNotFoundError:
            self._gcc_installed = False
            raise RuntimeError("Could not find gcc installation! Need gcc 9 or higher.")
        if cmd.returncode == 0:
            version = cmd.stdout.decode('UTF-8').strip().split('\n')[0]
            if int(version.split('.')[0]) < 9:
                self._gcc_installed = False
                raise RuntimeError(f"Need gcc 9 or higher, but found gcc {version}!")
            self._gcc_installed = True
            if verbose:
                cmd2 = run(["which", "gcc"], stdout=PIPE, stderr=PIPE)
                if cmd2.returncode == 0:
                    file = cmd2.stdout.decode('UTF-8').strip().split('\n')[0]
                    print(f"Found gcc {version} in {file}", flush=True)
                else:
                    stderr = cmd2.stderr.decode('UTF-8').strip().split('\n')
                    raise RuntimeError(f"Could not run `which gcc`!\nError given is:\n{stderr}")
        else:
            stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
            self._gcc_installed = False
            raise RuntimeError(f"Could not run gcc! Verify its installation.\nError given is:\n{stderr}")

    def assert_gfortran_installed(self, verbose=False):
        if hasattr(self, '_gfortran_installed'):
            return self._gfortran_installed
        try:
            cmd = run(["gfortran", "-dumpversion"], stdout=PIPE, stderr=PIPE)
        except FileNotFoundError:
            self._gfortran_installed = False
            raise RuntimeError("Could not find gfortran installation! Need gfortran 9 or higher.")
        if cmd.returncode == 0:
            version = cmd.stdout.decode('UTF-8').strip().split('\n')[0]
            if int(version.split('.')[0]) < 9:
                self._gfortran_installed = False
                raise RuntimeError(f"Need gfortran 9 or higher, but found gfortran {version}!")
            self._gfortran_installed = True
            if verbose:
                cmd2 = run(["which", "gfortran"], stdout=PIPE, stderr=PIPE)
                if cmd2.returncode == 0:
                    file = cmd2.stdout.decode('UTF-8').strip().split('\n')[0]
                    print(f"Found gfortran {version} in {file}", flush=True)
                else:
                    stderr = cmd2.stderr.decode('UTF-8').strip().split('\n')
                    raise RuntimeError(f"Could not run `which gfortran`!\nError given is:\n{stderr}")
        else:
            stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
            self._gfortran_installed = False
            raise RuntimeError(f"Could not run gfortran! Verify its installation.\nError given is:\n{stderr}")
